@@ -230,13 +230,23 @@ class GA:
 
         # Even such this parameter is declared in the class header, it is assigned to the object here to access it after saving the object.
         self.best_solutions_fitness = [] # A list holding the fitness value of the best solution for each generation.
-        
-        self.best_solution_generation = -1 # The generation number at which the best solution is reached. It is only assigned the generation number after the `run()` method completes. Otherwise, its value is -1.
+
+        self.best_solution_generation = -1 # The generation number at which the best fitness value is reached. It is only assigned the generation number after the `run()` method completes. Otherwise, its value is -1.
 
     def initialize_population(self, low, high):
+
         """
         Creates an initial population randomly as a NumPy array. The array is saved in the instance attribute named 'population'.
+
+        low: The lower value of the random range from which the gene values in the initial population are selected. It defaults to -4. Available in PyGAD 1.0.20 and higher.
+        high: The upper value of the random range from which the gene values in the initial population are selected. It defaults to -4. Available in PyGAD 1.0.20.
+        
+        This method assigns the values of the following 3 instance attributes:
+            1. pop_size: Size of the population.
+            2. population: Initially, holds the initial population and later updated after each generation.
+            3. init_population: Keeping the initial population.
         """
+
         # Population size = (number of chromosomes, number of genes per chromosome)
         self.pop_size = (self.sol_per_pop,self.num_genes) # The population will have sol_per_pop chromosome where each chromosome has num_genes genes.
         # Creating the initial population randomly.
@@ -247,10 +257,33 @@ class GA:
         # Keeping the initial population in the initial_population attribute.
         self.initial_population = self.population
 
+    def cal_pop_fitness(self):
+
+        """
+        Calculating the fitness values of all solutions in the current population. 
+        It returns:
+            -fitness: An array of the calculated fitness values.
+        """
+
+        if self.valid_parameters == False:
+            raise ValueError("ERROR calling the cal_pop_fitness() method: \nPlease check the parameters passed while creating an instance of the GA class.\n")
+
+        pop_fitness = []
+        # Calculating the fitness value of each solution in the current population.
+        for sol_idx, sol in enumerate(self.population):
+            fitness = self.fitness_func(sol, sol_idx)
+            pop_fitness.append(fitness)
+
+        pop_fitness = numpy.array(pop_fitness)
+
+        return pop_fitness
+
     def run(self):
+
         """
-        Running the genetic algorithm. This is the main method in which the genetic algorithm is evolved through a number of generations.
+        Runs the genetic algorithm. This is the main method in which the genetic algorithm is evolved through a number of generations.
         """
+
         if self.valid_parameters == False:
             raise ValueError("ERROR calling the run() method: \nThe run() method cannot be executed with invalid parameters. Please check the parameters passed while creating an instance of the GA class.\n")
 
@@ -292,34 +325,16 @@ class GA:
         # After the run() method completes, the run_completed flag is changed from False to True.
         self.run_completed = True # Set to True only after the run() method completes gracefully.
 
-    def cal_pop_fitness(self):
-        """
-        Calculating the fitness values of all solutions in the current population. 
-        It returns:
-            -fitness: An array of the calculated fitness values.
-        """
-        if self.valid_parameters == False:
-            raise ValueError("ERROR calling the cal_pop_fitness() method: \nPlease check the parameters passed while creating an instance of the GA class.\n")
-
-        pop_fitness = []
-        # Calculating the fitness value of each solution in the current population.
-        for sol_idx, sol in enumerate(self.population):
-            fitness = self.fitness_func(sol, sol_idx)
-            pop_fitness.append(fitness)
-
-        pop_fitness = numpy.array(pop_fitness)
-
-        return pop_fitness
-
     def steady_state_selection(self, fitness, num_parents):
+
         """
-        Applies the steady state selection of the parents that will mate to produce the offspring. 
+        Selects the parents using the steady-state selection technique. Later, these parents will mate to produce the offspring.
         It accepts 2 parameters:
             -fitness: The fitness values of the solutions in the current population.
             -num_parents: The number of parents to be selected.
-        It returns:
-            -parents: The selected parents to mate.
+        It returns an array of the selected parents.
         """
+
         fitness_sorted = sorted(range(len(fitness)), key=lambda k: fitness[k])
         fitness_sorted.reverse()
         # Selecting the best individuals in the current generation as parents for producing the offspring of the next generation.
@@ -329,14 +344,15 @@ class GA:
         return parents
 
     def rank_selection(self, fitness, num_parents):
+
         """
-        Applies the rank selection of the parents that will mate to produce the offspring. 
+        Selects the parents using the rank selection technique. Later, these parents will mate to produce the offspring.
         It accepts 2 parameters:
             -fitness: The fitness values of the solutions in the current population.
             -num_parents: The number of parents to be selected.
-        It returns:
-            -parents: The selected parents to mate.
+        It returns an array of the selected parents.
         """
+
         fitness_sorted = sorted(range(len(fitness)), key=lambda k: fitness[k])
         fitness_sorted.reverse()
         # Selecting the best individuals in the current generation as parents for producing the offspring of the next generation.
@@ -346,14 +362,15 @@ class GA:
         return parents
 
     def random_selection(self, fitness, num_parents):
+
         """
-        Randomly selecting the parents that will mate to produce the offspring. 
+        Selects the parents randomly. Later, these parents will mate to produce the offspring.
         It accepts 2 parameters:
             -fitness: The fitness values of the solutions in the current population.
             -num_parents: The number of parents to be selected.
-        It returns:
-            -parents: The selected parents to mate.
+        It returns an array of the selected parents.
         """
+
         parents = numpy.empty((num_parents, self.population.shape[1]))
 
         rand_indices = numpy.random.randint(low=0.0, high=fitness.shape[0], size=num_parents)
@@ -363,14 +380,15 @@ class GA:
         return parents
 
     def tournament_selection(self, fitness, num_parents):
+
         """
-        Applies the tournament selection of the parents that will mate to produce the offspring. 
+        Selects the parents using the tournament selection technique. Later, these parents will mate to produce the offspring.
         It accepts 2 parameters:
             -fitness: The fitness values of the solutions in the current population.
             -num_parents: The number of parents to be selected.
-        It returns:
-            -parents: The selected parents to mate.
+        It returns an array of the selected parents.
         """
+
         parents = numpy.empty((num_parents, self.population.shape[1]))
         for parent_num in range(num_parents):
             rand_indices = numpy.random.randint(low=0.0, high=len(fitness), size=self.K_tournament)
@@ -380,14 +398,15 @@ class GA:
         return parents
 
     def roulette_wheel_selection(self, fitness, num_parents):
+
         """
-        Applies the roulette wheel selection of the parents that will mate to produce the offspring. 
+        Selects the parents using the roulette wheel selection technique. Later, these parents will mate to produce the offspring.
         It accepts 2 parameters:
             -fitness: The fitness values of the solutions in the current population.
             -num_parents: The number of parents to be selected.
-        It returns:
-            -parents: The selected parents to mate.
+        It returns an array of the selected parents.
         """
+
         fitness_sum = numpy.sum(fitness)
         probs = fitness / fitness_sum
         probs_start = numpy.zeros(probs.shape, dtype=numpy.float) # An array holding the start values of the ranges of probabilities.
@@ -414,18 +433,15 @@ class GA:
         return parents
 
     def stochastic_universal_selection(self, fitness, num_parents):
+
         """
-        Applies the stochastic universal selection of the parents that will mate to produce the offspring. 
+        Selects the parents using the stochastic universal selection technique. Later, these parents will mate to produce the offspring.
         It accepts 2 parameters:
             -fitness: The fitness values of the solutions in the current population.
             -num_parents: The number of parents to be selected.
-        It returns:
-            -parents: The selected parents to mate.
+        It returns an array of the selected parents.
         """
-        # https://en.wikipedia.org/wiki/Stochastic_universal_sampling
-        # https://books.google.com.eg/books?id=gwUwIEPqk30C&pg=PA60&lpg=PA60&dq=Roulette+Wheel+genetic+algorithm+select+more+than+once&source=bl&ots=GLr2DrPcj4&sig=ACfU3U0jVOGXhzsla8mVqhi5x1giPRL4ew&hl=en&sa=X&ved=2ahUKEwim25rMvdzoAhWa8uAKHbt0AdgQ6AEwA3oECAYQLQ#v=onepage&q=Roulette%20Wheel%20&f=false
-        # https://www.tutorialspoint.com/genetic_algorithms/genetic_algorithms_parent_selection.htm
-        # https://www.obitko.com/tutorials/genetic-algorithms/selection.php
+
         fitness_sum = numpy.sum(fitness)
         probs = fitness / fitness_sum
         probs_start = numpy.zeros(probs.shape, dtype=numpy.float) # An array holding the start values of the ranges of probabilities.
@@ -455,14 +471,15 @@ class GA:
         return parents
 
     def single_point_crossover(self, parents, offspring_size):
+
         """
-        Applies the single point crossover. It selects a point randomly at which crossover takes place between two parents.
+        Applies the single-point crossover. It selects a point randomly at which crossover takes place between the pairs of parents.
         It accepts 2 parameters:
-            -parents: The parents to mate and create the offspring.
+            -parents: The parents to mate for producing the offspring.
             -offspring_size: The size of the offspring to produce.
-        It returns:
-            -offspring: The produced offspring after the parents mate.
+        It returns an array the produced offspring.
         """
+
         offspring = numpy.empty(offspring_size)
         # The point at which crossover takes place between two parents. Usually, it is at the center.
         crossover_point = numpy.random.randint(low=0, high=parents.shape[1], size=1)[0]
@@ -479,14 +496,15 @@ class GA:
         return offspring
 
     def two_points_crossover(self, parents, offspring_size):
+
         """
-        Applies the 2 points crossover. It selects the 2 points randomly at which crossover takes place between two parents.
+        Applies the 2 points crossover. It selects the 2 points randomly at which crossover takes place between the pairs of parents.
         It accepts 2 parameters:
-            -parents: The parents to mate and create the offspring.
+            -parents: The parents to mate for producing the offspring.
             -offspring_size: The size of the offspring to produce.
-        It returns:
-            -offspring: The produced offspring after the parents mate.
+        It returns an array the produced offspring.
         """
+
         offspring = numpy.empty(offspring_size)
         if (parents.shape[1] == 1): # If the chromosome has only a single gene. In this case, this gene is copied from the second parent.
             crossover_point1 = 0
@@ -509,14 +527,15 @@ class GA:
         return offspring
 
     def uniform_crossover(self, parents, offspring_size):
+
         """
         Applies the uniform crossover. For each gene, a parent out of the 2 mating parents is selected randomly and the gene is copied from it.
         It accepts 2 parameters:
-            -parents: The parents to mate and create the offspring.
+            -parents: The parents to mate for producing the offspring.
             -offspring_size: The size of the offspring to produce.
-        It returns:
-            -offspring: The produced offspring after the parents mate.
+        It returns an array the produced offspring.
         """
+
         offspring = numpy.empty(offspring_size)
 
         for k in range(offspring_size[0]):
@@ -536,13 +555,14 @@ class GA:
         return offspring
 
     def random_mutation(self, offspring):
+
         """
         Applies the random mutation which changes the values of a number of genes randomly by selecting a random value between random_mutation_min_val and random_mutation_max_val to be added to the selected genes.
         It accepts a single parameter:
             -offspring: The offspring to mutate.
-        It returns:
-            -offspring: The offspring after mutation.
+        It returns an array of the mutated offspring.
         """
+
         if self.mutation_num_genes == None:
             self.mutation_num_genes = numpy.uint32((self.mutation_percent_genes*offspring.shape[1])/100)
             # Based on the percentage of genes, if the number of selected genes for mutation is less than the least possible value which is 1, then the number will be set to 1.
@@ -550,20 +570,22 @@ class GA:
                 self.mutation_num_genes = 1
         mutation_indices = numpy.array(random.sample(range(0, offspring.shape[1]), self.mutation_num_genes))
         # Random mutation changes a single gene in each offspring randomly.
-        for idx in range(offspring.shape[0]):
-            # The random value to be added to the gene.
-            random_value = numpy.random.uniform(self.random_mutation_min_val, self.random_mutation_max_val, 1)
-            offspring[idx, mutation_indices] = offspring[idx, mutation_indices] + random_value
+        for offspring_idx in range(offspring.shape[0]):
+            for gene_idx in mutation_indices:
+                # The random value to be added to the gene.
+                random_value = numpy.random.uniform(self.random_mutation_min_val, self.random_mutation_max_val, 1)
+                offspring[offspring_idx, gene_idx] = offspring[offspring_idx, gene_idx] + random_value
         return offspring
 
     def swap_mutation(self, offspring):
+
         """
-        Applies the swap mutation which selects interchanges the values of 2 randomly selected genes.
+        Applies the swap mutation which interchanges the values of 2 randomly selected genes.
         It accepts a single parameter:
             -offspring: The offspring to mutate.
-        It returns:
-            -offspring: The offspring after mutation.
+        It returns an array of the mutated offspring.
         """
+
         for idx in range(offspring.shape[0]):
             mutation_gene1 = numpy.random.randint(low=0, high=offspring.shape[1]/2, size=1)[0]
             mutation_gene2 = mutation_gene1 + int(offspring.shape[1]/2)
@@ -574,13 +596,14 @@ class GA:
         return offspring
 
     def inversion_mutation(self, offspring):
+
         """
         Applies the inversion mutation which selects a subset of genes and invert them.
         It accepts a single parameter:
             -offspring: The offspring to mutate.
-        It returns:
-            -offspring: The offspring after mutation.
+        It returns an array of the mutated offspring.
         """
+
         for idx in range(offspring.shape[0]):
             mutation_gene1 = numpy.random.randint(low=0, high=numpy.ceil(offspring.shape[1]/2 + 1), size=1)[0]
             mutation_gene2 = mutation_gene1 + int(offspring.shape[1]/2)
@@ -590,13 +613,14 @@ class GA:
         return offspring
 
     def scramble_mutation(self, offspring):
+
         """
         Applies the scramble mutation which selects a subset of genes and shuffles their order randomly.
         It accepts a single parameter:
             -offspring: The offspring to mutate.
-        It returns:
-            -offspring: The offspring after mutation.
+        It returns an array of the mutated offspring.
         """
+
         for idx in range(offspring.shape[0]):
             mutation_gene1 = numpy.random.randint(low=0, high=numpy.ceil(offspring.shape[1]/2 + 1), size=1)[0]
             mutation_gene2 = mutation_gene1 + int(offspring.shape[1]/2)
@@ -608,8 +632,9 @@ class GA:
         return offspring
 
     def best_solution(self):
+
         """
-        Calculates the fitness values for the current population. 
+        Returns information about the best solution found by the genetic algorithm. Can only be called after completing at least 1 generation.
         If no generation is completed (at least 1), an exception is raised. Otherwise, the following is returned:
             -best_solution: Best solution in the current population.
             -best_solution_fitness: Fitness value of the best solution.
@@ -634,10 +659,10 @@ class GA:
         return best_solution, best_solution_fitness, best_match_idx
 
     def plot_result(self):
+
         """
-        Creating 2 plots that summarizes how the solutions evolved. Can only be called after completing at least 1 generation.
-        The first plot is between the iteration number and the function output based on the current parameters for the best solution.
-        The second plot is between the iteration number and the fitness value of the best solution.
+        Creates and shows a plot that summarizes how the fitness value evolved by generation. Can only be called after completing at least 1 generation.
+        If no generation is completed, an exception is raised.
         """
 
         if self.generations_completed < 1:
@@ -654,19 +679,23 @@ class GA:
         matplotlib.pyplot.show()
 
     def save(self, filename):
+
         """
-        Saving the genetic algorithm instance:
-            -filename: Name of the file to save the instance. It must have no extension.
+        Saves the genetic algorithm instance:
+            -filename: Name of the file to save the instance. No extension is needed.
         """
+
         with open(filename + ".pkl", 'wb') as file:
             pickle.dump(self, file)
 
 def load(filename):
+
     """
-    Reading a saved instance of the genetic algorithm:
-        -filename: Name of the file to read the instance. It must have no extension.
+    Reads a saved instance of the genetic algorithm:
+        -filename: Name of the file to read the instance. No extension is needed.
     Returns the genetic algorithm instance.
     """
+
     try:
         with open(filename + ".pkl", 'rb') as file:
             ga_in = pickle.load(file)
